/*
   (C) Sep 1998 Yoshiaki Shinmi 
   Author: Yoshiaki Shinmi 
              shinmi@ueda.info.waseda.ac.jp
*/

import java.io.*;
import java.net.*;
import java.lang.*;
import java.awt.*;
import java.awt.event.*;
import java.applet.Applet;
import java.util.Locale;

public class DcntlClient extends Applet {
    Socket socket = null;
    PrintWriter os = null;
    BufferedReader is = null;
    boolean do_process = true;
    boolean read_fail = false;

    Frame f = null;

    MenuPanel menu_p;
    InterfacePanel interface_p;
    StatusPanel status_p;
    QueuePanel queue_p;

    static String cmd_args[];

    boolean inAnApplet = true;
    String hostname;
    int port;
    String fsm;
    String up;
    String force;
    String impmode;
    String imp_itime;
    String imp_time;
    String imp_fuzz;
    String imp_timeout;
    String force_timeout;
    String timeout;
    String queue;
    String iface;
    String lip;
    String rip;
    int rx_load5;
    int tx_load5;
    int rx_load150;
    int tx_load150;
    int rx_total;
    int tx_total;

    public static void main(String args[]) {
	if (args.length != 2) {
	    System.out.println("DcntlClient:\n    Usage: java DcntlClient ServerAddress Port#");
	    System.exit(1);
	}
	
	cmd_args = args;

	DcntlClient dcntl = new DcntlClient();
	Frame f = new Frame("Diald Control");
	dcntl.f = f;

	dcntl.inAnApplet = false;
	dcntl.init();

	f.add("Center", dcntl);
	f.pack();
	f.setVisible(true);

	dcntl.start();
    }

    public void init() {

	if (inAnApplet == false) { 
	    try {
		port = Integer.parseInt(cmd_args[0]);
	    } catch (NumberFormatException e) {
		port = 8972;
		System.out.println("Port is invalid. Trying default port " + port);
	    }
	} else {
	    String hostStr, portStr;
	    hostStr = getParameter("SERVER_ADDR");
	    portStr = getParameter("PORT_NUM");
	    
	    if (hostStr == null) {
		hostname = getCodeBase().getHost();
	    } else {
		hostname = hostStr;
	    }
	    if (portStr == null) {
		port = 8972;
	    } else {
		try {
		    port = Integer.parseInt(portStr);
		} catch (NumberFormatException e) {
		    port = 8972;
		    System.out.println("Port is invalid. Trying default port "+ port);
		}
	    }
	}
	    
	setBackground(new Color(198,198,198));

	GridBagLayout gridbag = new GridBagLayout();
	GridBagConstraints c = new GridBagConstraints();
	setLayout(gridbag);

	menu_p = new MenuPanel(this);
	interface_p = new InterfacePanel(this);
	status_p = new StatusPanel(this);
	queue_p = new QueuePanel(this);

	c.fill = GridBagConstraints.BOTH;
	c.weightx = 1.0;
	c.weighty = 0.0;
	c.gridwidth = GridBagConstraints.REMAINDER;
	gridbag.setConstraints(menu_p, c);
	add(menu_p);
	c.gridwidth = GridBagConstraints.REMAINDER;
	gridbag.setConstraints(interface_p, c);
	add(interface_p);
	c.gridwidth = GridBagConstraints.REMAINDER;
	gridbag.setConstraints(status_p, c);
	add(status_p);
	c.weighty = 1.0;
	c.gridwidth = GridBagConstraints.REMAINDER;
	gridbag.setConstraints(queue_p, c);
	add(queue_p);

	setVisible(true);
	setEnabled(true);
    }

    public void start() {
	setSize(status_p.getPreferredSize());
	read_fail = false;
	do_process = true;
	setSize(600,300);
	processSocket();
    }

    private String readLine() {
	String buf;

	try {
	    buf = is.readLine();
	} catch (IOException e) {
	    System.out.println("readLine() error:" + e);
	    buf = null;
	    read_fail = true;
	}
	return buf;
    }

    private void processSocket() {
	String buf;

	if (socket == null) {
	    try {
		socket = new Socket(hostname, port);
	    } catch (UnknownHostException e) {
		System.err.println(e);
		stop();
	    } catch (IOException e) {
		System.err.println(e);
		stop();
	    } 
	}

	if (socket != null) {
	    if (os == null) {
		try {
		    // true: autoFlush
		    os = new PrintWriter(socket.getOutputStream(), true); 
		} catch (IOException e) {
		    System.err.println("DataOutputStream() failed:" + e);
		    stop();
		}
	    }
	    if (is == null) {
		try {
		    is = new BufferedReader(new InputStreamReader(socket.getInputStream()));
		} catch (IOException e) {
		    System.err.println("DataOutputStream() failed:" + e);
		    stop();
		}
	    }	  
	}

	buf = readLine();
	while(do_process) {
	    if (buf.equals("STATE")) {
		fsm = readLine();
		if (fsm.equals("CONNECT")) {
		    // change panel fsm
		}
		if (fsm.equals("UP")) {
		    // load init 
		}
		interface_p.setValues();
		interface_p.repaint();
		repaint();
	    }
	    if (buf.equals("STATUS")) {
		up = readLine();
		force = readLine();
		impmode = readLine();
		imp_itime = readLine();
		imp_time  = readLine();
		imp_fuzz = readLine();
		imp_timeout = readLine();
		force_timeout = readLine();
		timeout = readLine();

		menu_p.verifyState();
		status_p.setValues();
		status_p.repaint();
		repaint();
	    }
	    if (buf.equals("QUEUE")) {
		StringBuffer queueBuf = new StringBuffer();
		buf = readLine();
		while(!buf.equals("END QUEUE")) {
		    queueBuf.append(buf);
		    queueBuf.append("\n");
		    buf = readLine();
		}
		queue = queueBuf.toString();
		queue_p.setValues();
		queue_p.repaint();
		repaint();
	    }
	    if (buf.equals("MESSAGE")) {
		//
	    }
	    if (buf.equals("INTERFACE")) {
		iface = readLine();
		lip = readLine();
		rip = readLine();
		interface_p.setValues();
		interface_p.repaint();
		repaint();
	    }
	    if (buf.equals("LOAD")) {
		int tx, rx;
		tx = new Integer(readLine()).intValue();
		rx = new Integer(readLine()).intValue();
		rx_load5 = rx;
		tx_load5 = tx;
		rx_load150 = rx;
		tx_load150 = tx;
		rx_total += rx;
		tx_total += tx;
	    }
	    if (read_fail == true) {
		// socket is broken 
		System.out.println("Connection closed.\n");
		stop();
	    }
	    buf = readLine();
	}
    }
  
    public void stop() {
	do_process = false;

	try {
	    if (os != null) {
		os.close();
		os = null;
	    }
	    if (is != null) {
		is.close();
		is = null;
	    }
	    if (socket != null) { 
		socket.close();
		socket = null;
	    }
	} catch(IOException e) {
	    System.err.println(e);
	} 
    }

    public void send(String str) {
	if (os != null) {
	    os.write(str);
	    if (os.checkError() == true) { // checkError() do flush
		System.err.println("failed to write " + str);
		stop();
	    }
	}
    }

    public void paint(Graphics g) {
	Dimension d = getSize();
	if ((getPreferredSize().width > d.width) ||	
	    (getPreferredSize().height > d.height)) {
	    if (getPreferredSize().width > d.width)
		d.width = getPreferredSize().width;
	    if (getPreferredSize().height > d.height) 
		d.height = getPreferredSize().height;

	    setSize(d);

	    doLayout();
	    menu_p.doLayout();
	    interface_p.doLayout();
	    status_p.doLayout();
	    queue_p.doLayout();

	    if (f != null) {
		Insets i = f.getInsets();
		d.width += i.left + i.right;
		d.height += i.top + i.bottom;
		f.setSize(d);
	    }
	}
    }
}

class StatusPanel extends Panel {
    Label link_stat_l, link_stat_v;
    Label timeout_l,   timeout_v;
    Label force_l,     force_v;
    Label force_timeout_l, force_timeout_v;
    Label im_stat_l, im_stat_v;
    Label imp_itime_l, imp_itime_v;
    Label imp_time_l, imp_time_v;
    Label imp_fuzz_l, imp_fuzz_v;
    Label imp_timeout_l, imp_timeout_v;
    DcntlClient dcntlc;

    public StatusPanel(DcntlClient dcntlc) {
	this.dcntlc = dcntlc;

	link_stat_l = new Label("Link Status");
	link_stat_v = new Label();
	timeout_l = new Label("Next Alarm");
	timeout_v = new Label("00:00:00");
	force_l = new Label("Forcing Rule");
	force_v = new Label();
	force_timeout_l = new Label("Forcing Timeout");
	force_timeout_v = new Label("00:00:00");
	im_stat_l = new Label("Impulse State");
	im_stat_v = new Label();
	imp_itime_l = new Label("Initial Impulse");
	imp_itime_v = new Label();
	imp_time_l = new Label("Impulse Length");
	imp_time_v = new Label();
	imp_fuzz_l = new Label("Impulse Fuzz");
	imp_fuzz_v = new Label();
	imp_timeout_l = new Label("Impulse Timeout");
	imp_timeout_v = new Label("00:00:00");

	GridBagLayout gridbag = new GridBagLayout();
	GridBagConstraints c = new GridBagConstraints();
	setLayout(gridbag);

	// common constraints
	c.fill = GridBagConstraints.HORIZONTAL;
	c.weightx = 1.0;

	// first row
	gridbag.setConstraints(link_stat_l, c);
	add(link_stat_l);

	gridbag.setConstraints(link_stat_v, c);
	add(link_stat_v);

	gridbag.setConstraints(force_timeout_l, c);
	add(force_timeout_l);

	gridbag.setConstraints(force_timeout_v, c);
	add(force_timeout_v);

	gridbag.setConstraints(imp_time_l, c);
	add(imp_time_l);

	c.gridwidth = GridBagConstraints.REMAINDER; // end row
	gridbag.setConstraints(imp_time_v, c);
	add(imp_time_v);

	// second row
	c.gridwidth = 1;
	gridbag.setConstraints(timeout_l, c);
	add(timeout_l);
	gridbag.setConstraints(timeout_v, c);
	add(timeout_v);
	gridbag.setConstraints(im_stat_l, c);
	add(im_stat_l);
	gridbag.setConstraints(im_stat_v, c);
	add(im_stat_v);
	gridbag.setConstraints(imp_fuzz_l, c);
	add(imp_fuzz_l);
	c.gridwidth = GridBagConstraints.REMAINDER; // end row
	gridbag.setConstraints(imp_fuzz_v, c);
	add(imp_fuzz_v);

	// third row
	c.gridwidth = 1;
	gridbag.setConstraints(force_l, c);
	add(force_l);
	gridbag.setConstraints(force_v, c);
	add(force_v);
	gridbag.setConstraints(imp_itime_l, c);
	add(imp_itime_l);
	gridbag.setConstraints(imp_itime_v, c);
	add(imp_itime_v);
	gridbag.setConstraints(imp_timeout_l, c);
	add(imp_timeout_l);
	c.gridwidth = GridBagConstraints.REMAINDER; // end row
	gridbag.setConstraints(imp_timeout_v, c);
	add(imp_timeout_v);
    }
    public void setValues() {
	link_stat_v.setText(this.dcntlc.up);
	timeout_v.setText(this.dcntlc.timeout);
	force_v.setText(this.dcntlc.force);
	force_timeout_v.setText(this.dcntlc.force_timeout);
	im_stat_v.setText(this.dcntlc.impmode);
	imp_itime_v.setText(this.dcntlc.imp_itime);
	imp_time_v.setText(this.dcntlc.imp_time);
	imp_fuzz_v.setText(this.dcntlc.imp_fuzz);
	imp_timeout_v.setText(this.dcntlc.imp_timeout);
	doLayout();
	this.dcntlc.doLayout();
    }
}

class InterfacePanel extends Panel {
    Label iface_l, iface_v;
    Label lip_l, lip_v;
    Label rip_l, rip_v;
    Label fsm_l, fsm_v;

    DcntlClient dcntlc;

    public InterfacePanel(DcntlClient dcntlc) {
	this.dcntlc = dcntlc;
	FlowLayout flow;

	iface_l = new Label("Interface");
	iface_v = new Label();
	lip_l = new Label("from");
	lip_v = new Label();
	rip_l = new Label("to");
	rip_v = new Label();
	fsm_l = new Label("in state");
	fsm_v = new Label();

	flow = new FlowLayout(FlowLayout.LEFT);
	flow.setHgap(0);
	setLayout(flow);

	add(iface_l);
	add(iface_v);
	add(lip_l);
	add(lip_v);
	add(rip_l);
	add(rip_v);
	add(fsm_l);
	add(fsm_v);
    }
  
    public void setValues() {
	iface_v.setText(this.dcntlc.iface);
	lip_v.setText(this.dcntlc.lip);
	rip_v.setText(this.dcntlc.rip);
	fsm_v.setText(this.dcntlc.fsm);
	iface_v.invalidate();
	lip_v.invalidate();
	rip_v.invalidate();
	fsm_v.invalidate();
	doLayout();
	invalidate();
    }

    public void paint(Graphics g) {
	if (dcntlc.f != null) {
	    Insets i = dcntlc.f.getInsets();
	    Dimension d = dcntlc.f.getSize();
	    if (getPreferredSize().width > d.width - (i.left+i.right))  {
		d = dcntlc.getSize();
		d.width--;
		dcntlc.setSize(d);
		dcntlc.repaint();
	    } 
	}
	doLayout();
	dcntlc.doLayout();
    }
}

class QueuePanel extends Panel {
    Label title_l;
    TextArea queue_ta;

    DcntlClient dcntlc;

    public QueuePanel(DcntlClient dcntlc) {
	this.dcntlc = dcntlc;

	setLayout(new BorderLayout());
	title_l = new Label("Connection Queue");
	if (dcntlc.inAnApplet == true) {
	    queue_ta = new TextArea(12,40);
	} else {
	    queue_ta = new TextArea(7,40);
	}
	queue_ta.setEditable(false);

	add("North", title_l);
	add("Center", queue_ta);
    }

    public void setValues() {
	if (dcntlc.queue != null) {
	    queue_ta.setText(dcntlc.queue);
	}
    }
}


class MenuPanel extends Panel implements ActionListener, ItemListener {
    Button up_b, down_b;
    Checkbox force_cb, block_cb;
    DcntlClient dcntlc;

    String prev_state_str = new String();
    Integer oInt = new Integer(0);
    int force_state = 0;

    public MenuPanel(DcntlClient dcntlc) {
	this.dcntlc = dcntlc;
    
	setLayout(new FlowLayout(FlowLayout.LEFT));

	up_b = new Button("Up");
	up_b.setActionCommand("UP_BUTTON");
	up_b.addActionListener(this);

	down_b = new Button("Down");
	down_b.setActionCommand("DOWN_BUTTON");
	down_b.addActionListener(this);

	force_cb = new Checkbox("Forced up");
	force_cb.addItemListener(this);

	block_cb = new Checkbox("Block connection");  
	block_cb.addItemListener(this);

	add(up_b);
	add(down_b);
	add(force_cb);
	add(block_cb);
    }

    public void verifyState() {
	if ((dcntlc.force != null)
	    && (! prev_state_str.equals(dcntlc.force))) {

	    try {
		force_state = oInt.parseInt(dcntlc.force);
	    } catch (NumberFormatException e) {
		System.out.println("MenuPanel.verifyState():" + e); 
	    }
	    if (force_state % 2 == 1) {
		// Forced up is set
		force_cb.setState(true);
	    } else {
		// Forced up is unset
		force_cb.setState(false);
	    }
	    if (force_state / 2 == 1) {
		// Block connection is set
		block_cb.setState(true);
	    } else {
		// Block connection is unset
		block_cb.setState(false);
	    }
	    prev_state_str = new String(dcntlc.force);
	}
	//    if (dcntlc.force != null) prev_state_str = new String(dcntlc.force);
    }

    public void actionPerformed(ActionEvent e) {
	String cmd = e.getActionCommand();
	if (cmd.equals("UP_BUTTON")) {
	    dcntlc.send("up\n");
	} 
	if (cmd.equals("DOWN_BUTTON")) {
	    dcntlc.send("down\n");
	}
    }    

    public void itemStateChanged(ItemEvent e) {
	if (force_cb.equals(e.getSource())) {
	    if (e.getStateChange() == ItemEvent.SELECTED) {
		dcntlc.send("force\n");
	    } else {
		dcntlc.send("unforce\n");
	    }
	}
	if (block_cb.equals(e.getSource())) {
	    if (e.getStateChange() == ItemEvent.SELECTED) {
		dcntlc.send("block\n");
	    } else {
		dcntlc.send("unblock\n");
	    }
	}	
    }	
}
