#include<stdio.h>
#include<sys/socket.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<sys/signal.h>
#include<sys/time.h>
#include<netinet/in.h>
#include<fcntl.h>
#include<unistd.h>
#include<errno.h>
#include<syslog.h>


#define SOCK_MAX 5        /* listen() backlog */ 
#define CLIENT_MAX 10     /* $B%/%i%$%"%s%H:GBg?t(B  */

#ifndef DEFAULT_PORT
#define DEFAULT_PORT     8972     /* $B%]!<%HHV9f(B */
#endif

#define UNUSED -1         

#define BUFSIZE 1024                /* buffer size for socket, fifo */ 

#define SELECT_TOUT_SEC 0L         /* 0 sec: select() timeout sec.*/
#define SELECT_TOUT_USEC 10000L    /* 10 ms: select() timeout microsec.*/

#define CONTROL_FIFO "/etc/diald/diald.ctl"
#define PIDFILE "/var/run/dcntls.pid"

int fd0, fd1;
char fn[256];

void handler();

int s[CLIENT_MAX+2];           /* socket descriptor */

main() {
  int client_num = 0;
  int len, n, msglen, port;
  char monfifo[256];
  char *buf, *buf2;
  int bufsize, buf2size;
  char *tmp;
  int backlog_len[CLIENT_MAX+1];
  char *backlog[CLIENT_MAX+1];

  int i, j;
  int toggle = 0;

  int pid;
  FILE *fp;

  struct stat sbuf;

  fd_set readfds;
  struct sockaddr_in saddr;
  struct sockaddr_in caddr;
  
  struct timeval tv;

  fd0 = fd1 = UNUSED;

  signal(SIGINT, handler);     /* close fd and s in termination */
  signal(SIGTERM, handler);     /* close fd and s in termination */
  signal(SIGKILL, handler);     /* close fd and s in termination */

  /* open syslog */
  openlog("dcntls", LOG_PID | LOG_NDELAY | LOG_PERROR,  LOG_LOCAL2);

  /* 
   * become daemon
   */
  if ((pid = fork()) < 0) {
    syslog(LOG_ERR, "Could not fork into background: %m");
    exit(1);
  }
  /* parent process is finished */
  if (pid != 0) exit(0);
 
  if ((fp = fopen(PIDFILE, "w")) != NULL) {
    fprintf(fp, "%d\n", getpid());
    fclose(fp);
  } else {
    syslog(LOG_ERR, "Unable to create run file %s: %m", PIDFILE);
  }
  /* 
   * end of becoming daemon 
   */

  bufsize = BUFSIZE;
  buf2size = BUFSIZE;
  if ((buf = (char *)malloc(sizeof(char)*bufsize)) == NULL) {
    syslog(LOG_ERR, "cannot malloc for %d", bufsize);
    handler();
  }
  if ((buf2 = (char *)malloc(sizeof(char)*buf2size)) == NULL) {
    syslog(LOG_ERR, "cannot malloc for %d", buf2size);
    handler();
  }

  /* FIFO initializing */
  /* monitor fifo filename */
  sprintf(monfifo, "%s%d", "/tmp/dcntls.", (int)getpid());
  strcpy(fn, monfifo);

  /* open diald-control fifo */
  if ((fd0 =  open(CONTROL_FIFO, O_WRONLY)) < 0) {
    syslog(LOG_ERR, "cannot open diald-control FIFO %s: %m", CONTROL_FIFO);
    handler();
  }

  /* make monitor FIFO */
  /* delete obsolete monitor fifo */
  if (stat(monfifo, &sbuf) == 0) {
    if (unlink(monfifo) < 0) {
      syslog(LOG_ERR, "cannot delete obsolete monitor FIFO %s: %m",monfifo);
      handler();
    }
  }
  if (mknod(monfifo, S_IFIFO|0277, 0) < 0) {
    syslog(LOG_ERR, "cannot make FIFO %s: %m", monfifo);
    handler();
  }
  if (chmod(monfifo, 0600) < 0) {
    syslog(LOG_ERR, "fail to chmod %s: %m", monfifo); 
    handler();
  }
  if ((fd1 = open(monfifo, O_RDWR)) < 0) {
    syslog(LOG_ERR, "cannot open FIFO %s: %m", monfifo);
    handler();
  }
  /* set monitor fifo to non-blocking IO */
  if (fcntl(fd1,F_SETFL,fcntl(fd1,F_GETFL)|O_NONBLOCK) < 0) {
    syslog(LOG_ERR, "cannot set monitor fifo to non-blocking IO: %m ");
    handler();
  }

  /* request diald to monitor monfifo */
  sprintf(buf, "monitor %s\n",monfifo);
  if (write(fd0, buf, strlen(buf)) < 0) {
    syslog(LOG_ERR, "fail to request diald to monitor %s: %m", buf);
    handler();
  }
  /* End of FIFO initializing */

  /* Socket initializing */
  port = DEFAULT_PORT;
  
  tv.tv_sec = SELECT_TOUT_SEC;
  tv.tv_usec = SELECT_TOUT_USEC;

  for(i = 0; i <= CLIENT_MAX+1; i++) {
    s[i] = UNUSED;
  }

  if ((s[0] = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    syslog(LOG_ERR, "cannot make socket: %m");
    handler();
  }
  
  bzero((char *)&saddr, sizeof(saddr));
  
  saddr.sin_family = AF_INET;
  saddr.sin_addr.s_addr = INADDR_ANY;
  saddr.sin_port = htons(port);
  
  if (bind(s[0], (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
    syslog(LOG_ERR, "cannot bind socket: %m");
    handler();
  }

  if ((listen(s[0], SOCK_MAX)) < 0) {
    syslog(LOG_ERR, "cannot listen socket: %m");
    handler();
  }
  client_num = 0;

  /* prepare backlog */
  for(i = 0; i <= CLIENT_MAX; i++) {
    backlog_len[i] = 0;
    backlog[i] = (char *)malloc(BUFSIZE*sizeof(char));
  }

  close(0);
  close(1);
  close(2);

  while(1) {
    /* FIFO $B$+$i(B $B%G!<%?$rFI$_9~$_!"(B $B%/%i%$%"%s%H$KAw?.(B  */
    /* FIFO$B$r(B 
       fcntl(fd,F_SETFL,fcntl(fd,F_GETFL)|O_NONBLOCK);
       $B$G%N%s%V%m%C%-%s%0$K@_Dj$7!"FI$_9~$_$G%V%m%C%/$7$J$$$h$&$K$9$k!#(B
       */
    /* ....... */
    usleep(100000); 
    
    len = 0;
    while ((msglen = read(fd1, buf2, buf2size*sizeof(char))) > 0) {
      while (bufsize <= len+msglen) {
	bufsize = bufsize * 2;
	tmp = (char *)realloc(buf, sizeof(char)*bufsize);
	if (tmp == NULL) {
	  syslog(LOG_ERR, "realloc failed for %d bytes",sizeof(char)*bufsize);
	  handler();
	} else {
	  buf = tmp;
	}
      }
      bcopy(buf2, &buf[len], msglen*sizeof(char));
      len += msglen;      
    }
    if (msglen <= 0 && errno != EAGAIN) {
      /* Fatal error 
       * cannot read from monitor fifo
       */
      syslog(LOG_ERR, "cannot read from monitor fifo: %m");
      handler();
    }


    /* 
     * send data to all clients */

    if (len > 0) {
      for(i = 1; i <= CLIENT_MAX; i++) {
	if (s[i] != UNUSED) {
	  if (backlog_len[i] > 0) {
	    bcopy(backlog[i], buf2, backlog_len[i]);
	    bcopy(buf, &buf2[backlog_len[i]], len);
	    len = len + backlog_len[i];
	    bcopy(buf2, buf, len);
	  }
	  msglen = write(s[i], buf, len);
	  if (msglen < 0) {
	    if (errno == EAGAIN) {
	      /* no data sended  but no problem */
	      syslog(LOG_INFO,"no data sended to client #%d", i);
	    } else {
	      /* fatal error */
	      syslog(LOG_ERR, "write error to client #%d: %m", i);
	    }
	  } else if (msglen == 0) {
	    /* socket closed */
	    syslog(LOG_INFO, "client #%d: connection closed.", i);
	    close(s[i]);
	    s[i] = UNUSED;
	    client_num--;
	  } else if (msglen < len) {
	    /* a part of data sended. must treat backlog next time */
	    if (msglen < backlog_len[i]) {
	      /* $BA0$N(Bbacklog $B$9$iAw$l$J$$$J$i!"$=$l$K$H$I$a$k(B */
	      syslog(LOG_INFO,"sended a part of backlog to client #%d. some are left", i);
	      backlog_len[i] = backlog_len[i] - msglen;
	      bcopy(&buf[msglen], backlog[i], backlog_len[i]);
	    } else {
	      /* $BA0$N(Bbacklog $B$OAw$l$?$+$i!"<!$N(Bbacklog$B$rMQ0U(B */
	      syslog(LOG_INFO,"backlog all sended to client #%d. but some are left", i);
	      backlog_len[i] = len - msglen;
	      bcopy(&buf[msglen], backlog[i], backlog_len[i]);
	    }
	  } else {
	    /* all data sended */
	    backlog_len[i] = 0;
	  }
	}
      }
    }
    /* $B%/%i%$%"%s%H$+$i$N%G!<%?FI$_9~$_$H?75,%/%i%$%"%s%H$N<uIU(B */
    FD_ZERO(&readfds);
    
    for(i = 0; i <= CLIENT_MAX; i++) {
      if (s[i] != UNUSED) {
	FD_SET(s[i], &readfds);
      }
    }
    
    if ((n = select(FD_SETSIZE, &readfds, NULL, NULL, &tv)) < 0) { 
      syslog(LOG_ERR, "select error on reading request from client: %m");
      handler();
    }
    
    for(i = 1; i <= CLIENT_MAX; i++) {
      if (s[i] != UNUSED) {
	if (FD_ISSET(s[i], &readfds)) {
	  /* 
	   * s[i]$B$N%S%C%H$,N)$C$F$$$l$P!"(Bs[i]$B$K%a%C%;!<%8$,E~Ce$7$F$$$k!#(B
	   */
	  bzero(buf2, buf2size);
	  if ((msglen = read(s[i], buf2, buf2size)) < 0) {
	    if (errno != EAGAIN) {
	      /* read failed */
	      syslog(LOG_ERR,"socket %d read error: %m", i);
	    }
	  } else if (msglen != 0) {
	    /* read succeeded */
	    /* send data to diald control fifo */
	    if (len = write(fd0, buf2, msglen) < 0) {
	      syslog(LOG_ERR, "cannot write control %s to diald: %m", buf2);
            }
	  } else {
	    syslog(LOG_INFO, "client #%d: connection closed.", i);
	    close(s[i]);
	    s[i] = UNUSED;
	    client_num--;
	  }
	}
      }
    }

    /* $B?75,%/%i%$%"%s%H$+$i$N@\B3MW5a$N>l9g$O!"(Bs[0]$B$N%S%C%H$,N)$D!#(B
       $B0J2<$G$O!"?75,%/%i%$%"%s%H$N@\B3<uIU$r9T$J$&!#(B
       */
    if (FD_ISSET(s[0], &readfds)) {
      /* $B6u$-%=%1%C%H$rC5$9(B */
      i = 1; 
      while(i <= CLIENT_MAX && s[i] != UNUSED) i++;
      
      /* $B@\B3$r3NN)(B accept() */
      len = sizeof(caddr);
      s[i] = accept(s[0], (struct sockaddr *)&caddr, &len);
      
      if (s[i] < 0) {
	syslog(LOG_ERR, "cannot accept client: %m");
	handler();
      }
      if (i > CLIENT_MAX) {
	/* $B%5!<%P!<%S%8!<(B */
	syslog(LOG_INFO, "Server busy: refuse connection.");
	strcpy(buf, "Server is busy. Try later!!\n");
	write(s[i], buf, strlen(buf)+1);
	close(s[i]);
	s[i] = UNUSED;
      } else {
	/* 
	 * new client accepting succeeded 
	 */
	syslog(LOG_INFO, "client #%d: accepted from %s\n", 
	i, inet_ntoa(caddr.sin_addr.s_addr));
	client_num++;

	/* set the socket to non-blocking */	
	if (fcntl(s[i],F_SETFL,fcntl(s[i],F_GETFL)|O_NONBLOCK) < 0) { 
	  syslog(LOG_ERR, "cannot set socket #%d to non-blocking: %m", i);
	}
	/* request diald to monitor monfifo to get current FSM state */
	sprintf(buf, "monitor %s\n",monfifo);
	if (write(fd0, buf, strlen(buf)) < 0) {
	  syslog(LOG_ERR, "fail to request diald to send current state: %m");
	}
      }
    }
  }
}
      
void handler() {
  int i;
  struct stat sbuf;

  if (fd0 != UNUSED) {
    close(fd0);
  }
  if (fd1 != UNUSED) {
    close(fd1);
    unlink(fn);
  }
  for(i = 0; i <= CLIENT_MAX+1; i++) { 
    if (s[i] != UNUSED) 
      close(s[i]);
  }
  if (stat(PIDFILE, &sbuf) == 0) {
    if (unlink(PIDFILE) < 0) {
      syslog(LOG_ERR, "cannot delete pidfile: %s",PIDFILE);
      exit(1);
    }
  }
  exit(0);
}

